/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains a doubly linked list.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include "list.h"
#include <stdlib.h>

/**
 * Create a new list.
 *
 * @return A new, empty list.
 */
extern list *list_create(){
    list *l;
    if((l = (list *)malloc(sizeof(list)))) {
        l->first   = NULL;
        l->last    = NULL;
        l->current = NULL;
    }
    return l;
}

/**
 * Remove the list from memory.
 *
 * @param l The list we should remove.
 */
extern void list_destroy(list *l){
    listelem *next = l->first;
    listelem *current;
    while(next) {
        current = next;
        next = next->next;
        free(current);
    }
    free(l);
}

/**
 * Checks to see if the list is empty.
 * @param l The list to test.
 * @return 1 if the list is empty, 0 otherwise.
 */
extern int list_is_empty(const list *const l){
    if(l->first) {
        return 0;
    }
    return 1;
}
 
/**
 * Sets "current" element to the next element.
 *
 * @param l The list to change.
 * @return The next element on success, NULL if such an element does not exist.
 */
extern void *list_next(list *const l){
    if(l->current && l->current->next) {
        l->current = l->current->next;
        return l->current->elem;
    }
    return NULL;
}
 
/**
 * Sets "current" element to the previous element.
 *
 * @param l The list to change.
 * @return The previous element on success, NULL if such an element does not
 * exist.
 */
extern void *list_prev(list *const l){
    if(l->current && l->current->prev) {
        l->current = l->current->prev;
        return l->current->elem;
    }
    return NULL;
}

/**
 * Sets "current" element to the first element.
 *
 * @param l The list to change.
 * @return The first element on success, NULL if such an element does not
 */
extern void *list_first(list *const l){
    l->current = l->first;
    if(l->current) {
        return l->current->elem;
    }
    return NULL;
}
 
/**
 * Sets "current" element to the last element.
 *
 * @param l The list to change.
 * @return The last element on success, NULL if such an element does not
 */
extern void *list_last(list *const l){
    l->current = l->last;
    if(l->current) {
        return l->current->elem;
    }
    return NULL;
}
 
/**
 * Retrieve the "current" element.
 *
 * @param l The list to get the element from.
 * @return The "current" element, or NULL if such an element does not exist.
 */
extern void *list_this(const list *const l){
    if(l->current) {
        return l->current->elem;
    }
    return NULL;
}

/**
 * Insert an element in the list.
 *
 * @param l The list to insert the element into.
 * @param elem The element to insert into the list.
 * @param where If 0 the element will be inserted after current element,
 * otherwise it is inserted before the current element.
 * @return 0 on on success, otherwise -1.
 */
extern int list_insert(list *const l, void *elem, const int where){
    listelem *new_listelem;
    if(!(new_listelem = (listelem *)malloc(sizeof(listelem)))) {
        return -1;
    }
    new_listelem->elem = elem;
 
    if(!l->current) { /* List is empty. */
        new_listelem->prev = NULL;
        new_listelem->next = NULL;
        l->first = l->last = l->current = new_listelem;
    } else {
        if(where == LIST_AFTER) { /* Put after current. */
            new_listelem->prev = l->current;
            if(!(new_listelem->next = l->current->next)) {
                /* New element is last. */
                l->last = new_listelem;
            } else {
                l->current->next->prev = new_listelem;
            }
            l->current->next = new_listelem;
        } else { /* Put before current. */
            new_listelem->next = l->current;
            if(!(new_listelem->prev = l->current->prev)) {
                /* New element is first. */
                l->first = new_listelem;
            } else {
                l->current->prev->next = new_listelem;
            }
            l->current->prev = new_listelem;
        }
        l->current = new_listelem;
    }
    return 0;
}
 
/**
 * Deletes "current" element. The function will not attempt to free the memory
 * used by the element.
 *
 * @param l The list to delete the element from.
 * @return A pointer to the data of the deleted node.
 */
extern void *list_delete(list *const l){
    void *data = NULL;
    listelem *old_listelem;
    if(!(old_listelem = l->current)) {
        return NULL; /* Empty list. */
    }
 
    if(old_listelem->next) { /* Deleted element is not last. */
        l->current = old_listelem->next;
        old_listelem->next->prev = old_listelem->prev;
    } else { /* Deleted element is last. */
        l->current = old_listelem->prev;
        l->last = old_listelem->prev;
    }
 
    if(old_listelem->prev) { /* Deleted element is not first. */
        old_listelem->prev->next = old_listelem->next;
    } else { /* Deleted element is first. */
        l->first = old_listelem->next;
    }

    data = old_listelem->elem;
    free(old_listelem);
    return data;
}

/**
 * Get the number of elements in the list.
 * @param l The list to count.
 * @return The number of elements in the list.
 */
extern int list_count(list *const l){
    int count = 1;
    listelem *elem = l->first;
    if(l->first == NULL) {
        return 0;
    }
    while((elem = elem->next)) {
        count++;
    }
    return count;
}

/**
 * Get the element at a specific position in the list and set the current
 * element to that position.
 *
 * @param l The list to work with.
 * @param index The 0-based index of the requested element.
 * @return The element at position index, or NULL if there is no element there.
 */
extern void *list_jump(list *const l, int index){
    listelem *e = NULL;
    
    if(l == NULL) {
        return NULL;
    }
    if((index < 0) || (index >= list_count(l))) {
        return NULL;
    }

    e = l->first;
    while(index > 0) {
        e = e->next;
        index--;
    }
    l->current = e;
    return e->elem;
}
