/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains a doubly linked list.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _list_h_
#define _list_h_

/* Constants for placement at insert. */
enum{
    LIST_AFTER  = 0,
    LIST_BEFORE = 1
};

typedef struct listelem listelem;
typedef struct list list;

/* The data type representing an element in a list. */
struct listelem{
    listelem *prev;
    listelem *next;
    void *elem;
};

/* The data type representing a list. */
struct list{
    listelem *first;
    listelem *last;
    listelem *current;
};

/**
 * Create a new list.
 *
 * @return A new, empty list.
 */
extern list *list_create();

/**
 * Remove the list from memory.
 *
 * @param l The list we should remove.
 */
extern void list_destroy(list *l);

/**
 * Checks to see if the list is empty.
 * @param l The list to test.
 * @return 1 if the list is empty, 0 otherwise.
 */
extern int list_is_empty(const list *const l);
 
/**
 * Sets "current" element to the next element.
 *
 * @param l The list to change.
 * @return The next element on success, NULL if such an element does not exist.
 */
extern void *list_next(list *const l);
 
/**
 * Sets "current" element to the previous element.
 *
 * @param l The list to change.
 * @return The previous element on success, NULL if such an element does not
 * exist.
 */
extern void *list_prev(list *const l);
 
/**
 * Sets "current" element to the first element.
 *
 * @param l The list to change.
 * @return The first element on success, NULL if such an element does not
 */
extern void *list_first(list *const l);
 
/**
 * Sets "current" element to the last element.
 *
 * @param l The list to change.
 * @return The last element on success, NULL if such an element does not
 */
extern void *list_last(list *const l);
 
/**
 * Retrieve the "current" element.
 *
 * @param l The list to get the element from.
 * @return The "current" element, or NULL if such an element does not exist.
 */
extern void *list_this(const list *const l);

/**
 * Insert an element in the list.
 *
 * @param l The list to insert the element into.
 * @param elem The element to insert into the list.
 * @param where If 0 the element will be inserted after current element,
 * otherwise it is inserted before the current element.
 * @return 0 on on success, otherwise -1.
 */
extern int list_insert(list *const l, void *elem, const int where);
 
/**
 * Deletes "current" element. The function will not attempt to free the memory
 * used by the element.
 *
 * @param l The list to delete the element from.
 * @return A pointer to the data of the deleted node.
 */
extern void *list_delete(list *const l);

/**
 * Get the number of elements in the list.
 *
 * @param l The list to count.
 * @return The number of elements in the list.
 */
extern int list_count(list *const l);

/**
 * Get the element at a specific position in the list and set the current
 * element to that position.
 *
 * @param l The list to work with.
 * @param index The 0-based index of the requested element.
 * @return The element at position index, or NULL if there is no element there.
 */
extern void *list_jump(list *const l, int index);

#endif
