/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the implementation of the libmadcap API.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include "madcap.h"
#include "message.h"
#include "option.h"
#include "net.h"
#include "list.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

/* The default local scope for IPv4. */
static name local_scope_name = {
    "en",
    "Default local scope"
};

static const scope local_scope = {
    {
        0xEFFF0000, /* 239.255.0.0 */
        0xEFFFFFFF  /* 239.255.255.255 */
    },
    &local_scope_name,
    1,
    16
};

#ifdef DEBUGGING
static void print_ecode(mc_msg *msg){
    int i = 0;
    fprintf(stderr, "Error code: %d\n",
            ((mc_option_error *)msg_get_option(msg, MC_ERROR))->ecode);
    fprintf(stderr, "Error data: ");
    for(i = 0;
        i < (((mc_option_error *)msg_get_option(msg, MC_ERROR))->len - 2);
        i++) {
        fprintf(stderr, " %02X ",
                (((mc_option_error *)msg_get_option(msg, MC_ERROR))->data)[i]);
    }
    fprintf(stderr, "\n");
}
#endif

/**
 * General function for allocation of multicast addresses.
 * This function is also called by the more specialized allocation functions.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param server The server from which to allocate the address.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
int alloc_multicast_addr(uint16 address_family,
                         scope *scope,
                         uint16 min_desired_addresses,
                         uint16 max_desired_addresses,
                         uint32 min_desired_start_time,
                         uint32 max_desired_start_time,
                         uint32 min_desired_lifetime,
                         uint32 max_desired_lifetime,
                         uint32 server,
                         lease *multicast_address_set_lease,
                         int *status){

    mc_msg              *request_msg   = NULL;
    mc_msg              *reply_msg     = NULL;
    mc_option_sid       *sid_opt       = NULL;
    mc_option_lid       *lid_opt       = NULL;
    mc_option_scope     *scope_opt     = NULL;
    mc_option_numaddr   *numaddr_opt   = NULL;
    mc_option_stime     *stime_opt     = NULL;
    mc_option_max_stime *max_stime_opt = NULL;
    mc_option_ltime     *ltime_opt     = NULL;
    mc_option_min_ltime *min_ltime_opt = NULL;
    mc_option_ctime     *ctime_opt     = NULL;

    uint8 *raw_message = NULL;
    int message_size = 0;
    int socket = -1;
    int timeout = -1;
    address_range *range = NULL;

    
    /* First check the arguments. */
    if(address_family != MC_IPV4) {
        mc_errno = MC_EFAMILY;
        DEBUG("Address family invalid in alloc_multicast_addr");
        return -1;
    }
    /* TODO: Better validity check for scope. */
    if(scope == NULL) {
        mc_errno = MC_ESCOPE;
        DEBUG("Problem with scope in alloc_multicast_addr");
        return -1;
    }
    if(max_desired_addresses < min_desired_addresses) {
        mc_errno = MC_ENUMADDR;
        DEBUG("Number of addresses invalid in alloc_multicast_addr");
        return -1;
    }
    if(max_desired_lifetime < min_desired_lifetime) {
        mc_errno = MC_ELTIME;
        DEBUG("Requested lifetime invalid in alloc_multicast_addr");
        return -1;
    }
    if(max_desired_start_time < min_desired_start_time) {
        mc_errno = MC_ESTIME;
        DEBUG("Requested start time invalid in alloc_multicast_addr");
        return -1;
    }

    sid_opt = option_create_sid(server);
    lid_opt = option_create_lid(NULL, 0);
    scope_opt = option_create_scope(scope->addresses.start);
    
    request_msg = msg_create_request(sid_opt, lid_opt, scope_opt);

    if(max_desired_addresses > 0) {
        /* We need a NUMADDR option. */
        numaddr_opt = option_create_numaddr(min_desired_addresses,
                                            max_desired_addresses);
        msg_add_option(request_msg, (mc_option *)numaddr_opt);
    }

    if(min_desired_start_time > 0) {
        /* We need the STIME option. */
        stime_opt = option_create_stime(min_desired_start_time);
        msg_add_option(request_msg, (mc_option *)stime_opt);
    }
    
    if(max_desired_start_time > 0) {
        /* We need the MAX_STIME option. */
        max_stime_opt = option_create_max_stime(max_desired_start_time);
        msg_add_option(request_msg, (mc_option *)max_stime_opt);
        /* STIME and MAX_STIME requires that CTIME is added. */
        ctime_opt = option_create_ctime(0);
        msg_add_option(request_msg, (mc_option *)ctime_opt);
    }

    if(max_desired_lifetime > 0) {
        /* We need the LTIME option. */
        ltime_opt = option_create_ltime(max_desired_lifetime);
        msg_add_option(request_msg, (mc_option *)ltime_opt);
    }

    if(min_desired_lifetime > 0) {
        /* We need the MIN_LTIME option. */
        min_ltime_opt = option_create_min_ltime(min_desired_lifetime);
        msg_add_option(request_msg, (mc_option *)min_ltime_opt);
    }

    /* Create and add the End option. */
    msg_add_option(request_msg, (mc_option *)option_create_end());

    /* Encode message. */
    raw_message = mc_encode_message(request_msg, &message_size);

    /* Send message. */
    socket = mc_send_message(raw_message, message_size, server, MC_SERVERPORT,
                             0);
    free(raw_message);
    raw_message = NULL;

    /* Receive reply. */
    timeout = MC_BASE_TIMEOUT;
    while((timeout <= 16) && (timeout <= (MC_BASE_TIMEOUT * 4))) {
        raw_message = mc_receive_message(socket, timeout, &message_size);
        if(raw_message == NULL) {
            if(mc_errno == MC_ETIMEOUT) {
                /* Time-out so we increase timeout and try again. */
                timeout = timeout * 2;
                continue;
            } else if(mc_errno == MC_EINTR) {
                /* Interrupted by something else then time-out. Try again. */
                continue;
            }
        }
        break;
    }

    if(raw_message == NULL) {
        DEBUG("Reception of message failed in alloc_multicast_addr.");
        msg_delete(request_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Decode message. */
    reply_msg = mc_decode_message(raw_message, message_size);
    free(raw_message);
    raw_message = NULL;
    if(reply_msg == NULL) {
        DEBUG("Failed to decode message in alloc_multicast_addr.");
        msg_delete(request_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Validate message. */
    if(msg_validate(reply_msg, request_msg) == -1) {
        msg_delete(request_msg);
        msg_delete(reply_msg);
        DEBUG("Failed to validate message in alloc_multicast_addr.");
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* No further need for the sent message. */
    msg_delete(request_msg);
    request_msg = NULL;

    if(reply_msg->msgtype == MC_NAK) {
        /* Easy case, we failed. */
#ifdef DEBUGGING
        print_ecode(reply_msg);
#endif
        msg_delete(reply_msg);
        DEBUG("Received NAK in alloc_multicast_address.");
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Here we construct the lease and status. */
    range = list_first(((mc_option_addrlist *)
                        msg_get_option(reply_msg, MC_ADDRLIST))->ranges);
    multicast_address_set_lease->addresses = NULL;
    multicast_address_set_lease->num_addr = 0;
    while(range != NULL) {
        uint32 tmp_addr = range->start;
        while(tmp_addr <= range->end) {
            multicast_address_set_lease->addresses = realloc(
                multicast_address_set_lease->addresses,
                sizeof(uint32) * (multicast_address_set_lease->num_addr + 1));
            ASSERT(multicast_address_set_lease->addresses != NULL,
                   "Memory allocation failed.");
            *(multicast_address_set_lease->addresses +
              multicast_address_set_lease->num_addr) = tmp_addr;
            multicast_address_set_lease->num_addr++;
            tmp_addr++;
        }
        range = list_next(((mc_option_addrlist *)
                           msg_get_option(reply_msg, MC_ADDRLIST))->ranges);
    }
    multicast_address_set_lease->address_family = reply_msg->addrfamily;
    multicast_address_set_lease->lifetime =
        ((mc_option_ltime *)msg_get_option(reply_msg, MC_LTIME))->lease_time;
    if(msg_get_option(reply_msg, MC_STIME)) {
        multicast_address_set_lease->start_time = 
            ((mc_option_stime *)msg_get_option(reply_msg, MC_STIME))
            ->start_time;
    } else {
        multicast_address_set_lease->start_time = time(NULL);
    }
    multicast_address_set_lease->lease_id =
        (uint8 *)malloc(((mc_option_lid *)msg_get_option(reply_msg, MC_LID))->
                        len * sizeof(uint8));
    ASSERT(multicast_address_set_lease->lease_id != NULL,
           "Memory allocation failed.");
    memcpy(multicast_address_set_lease->lease_id,
           ((mc_option_lid *)msg_get_option(reply_msg, MC_LID))->id,
           ((mc_option_lid *)msg_get_option(reply_msg, MC_LID))->len);
    multicast_address_set_lease->id_len =
        ((mc_option_lid *)msg_get_option(reply_msg, MC_LID))->len;
    multicast_address_set_lease->server_address =
        ((mc_option_sid *)msg_get_option(reply_msg, MC_SID))->address;

    *status = MC_SOK;

    msg_delete(reply_msg);
    reply_msg = NULL;
    
    return 0;
}

/**
 * General function for allocation of multicast addresses when the server is not
 * known.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
int alloc_noserv_addr(uint16 address_family,
                      scope *scope,
                      uint16 min_desired_addresses,
                      uint16 max_desired_addresses,
                      uint32 min_desired_start_time,
                      uint32 max_desired_start_time,
                      uint32 min_desired_lifetime,
                      uint32 max_desired_lifetime,
                      lease *multicast_address_set_lease,
                      int *status){
    mc_msg            *send_msg    = NULL;
    mc_msg            *reply_msg   = NULL;
    mc_option_lid     *lid_opt     = NULL;
    mc_option_reqlist *reqlist_opt = NULL;
    uint16            sid_id       = MC_SID;
    uint8             *raw_message = NULL;
    int               message_size = 0;
    int               socket       = -1;
    int               timeout      = -1;
    uint32            server       = 0;
    
    /* First create a GETINFO so we can find a server. */
    lid_opt = option_create_lid(NULL, 0);
    reqlist_opt = option_create_reqlist(1, &sid_id);
    
    send_msg = msg_create_getinfo(lid_opt, reqlist_opt);

    /* Create and add the End option. */
    msg_add_option(send_msg, (mc_option *)option_create_end());

    /* Encode message. */
    raw_message = mc_encode_message(send_msg, &message_size);

    /* Send message. */
    socket = mc_send_message(raw_message, message_size,
                             scope->addresses.end -1, MC_SERVERPORT,
                             local_scope.ttl);
    free(raw_message);
    raw_message = NULL;

    /* Receive reply.  Note that we only wait for one reply to arrive. The
     protocol specification allows this. */
    timeout = MC_BASE_TIMEOUT;
    while((timeout <= 16) && (timeout <= (MC_BASE_TIMEOUT * 4))) {
        raw_message = mc_receive_message(socket, timeout, &message_size);
        if(raw_message == NULL) {
            if(mc_errno == MC_ETIMEOUT) {
                /* Time-out so we increase timeout and try again. */
                timeout = timeout * 2;
                continue;
            } else if(mc_errno == MC_EINTR) {
                /* Interrupted by something else then time-out. Try again. */
                continue;
            }
        }
        break;
    }

    if(raw_message == NULL) {
        DEBUG("Reception of message failed in alloc_noserv_addr.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Decode message. */
    reply_msg = mc_decode_message(raw_message, message_size);
    free(raw_message);
    raw_message = NULL;
    if(reply_msg == NULL) {
        DEBUG("Failed to decode message in alloc_noserv_addr.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Validate message. */
    if(msg_validate(reply_msg, send_msg) == -1) {
        msg_delete(send_msg);
        msg_delete(reply_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* No further need for the sent message. */
    msg_delete(send_msg);
    send_msg = NULL;
    
    /* Now extract the address of the server which replied. */
    server = ((mc_option_sid *)msg_get_option(reply_msg, MC_SID))->address;

    /* Clean up. */
    msg_delete(reply_msg);
    reply_msg = NULL;

    /* And then it is time for the real allocation. */
    return alloc_multicast_addr(address_family,
                                scope,
                                min_desired_addresses,
                                max_desired_addresses,
                                min_desired_start_time,
                                max_desired_start_time,
                                min_desired_lifetime,
                                max_desired_lifetime,
                                server,
                                multicast_address_set_lease,
                                status);
}

/**
 * Allocate a single multicast address.
 * This function calles the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
int alloc_one_addr(uint16 address_family,
                   scope *scope,
                   uint32 min_desired_start_time,
                   uint32 max_desired_start_time,
                   uint32 min_desired_lifetime,
                   uint32 max_desired_lifetime,
                   lease *multicast_address_set_lease,
                   int *status){
    return alloc_noserv_addr(address_family,
                             scope,
                             1, /* min_desired_addresses */
                             1, /* max_desired_addresses */
                             min_desired_start_time,
                             max_desired_start_time,
                             min_desired_lifetime,
                             max_desired_lifetime,
                             multicast_address_set_lease,
                             status);
}

/**
 * Function for allocation of addresses with an exact lifetime.
 * This function will call the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param lifetime The time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
int alloc_exact_addr(uint16 address_family,
                     scope *scope,
                     unsigned int min_desired_addresses,
                     unsigned int max_desired_addresses,
                     uint32 min_desired_start_time,
                     uint32 max_desired_start_time,
                     uint32 lifetime,
                     lease *multicast_address_set_lease,
                     int *status){
    return alloc_noserv_addr(address_family,
                             scope,
                             min_desired_addresses,
                             max_desired_addresses,
                             min_desired_start_time,
                             max_desired_start_time,
                             lifetime, /* min_desired_lifetime */
                             lifetime, /* max_desired_lifetime */
                             multicast_address_set_lease,
                             status);
}

/**
 * Function for allocation of addresses which starts immediately.
 * This function will call the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
int alloc_immediate_addr(uint16 address_family,
                         scope *scope,
                         int min_desired_addresses,
                         int max_desired_addresses,
                         uint32 min_desired_lifetime,
                         uint32 max_desired_lifetime,
                         lease *multicast_address_set_lease,
                         int *status){
    uint32 time_now = time(NULL);
    return alloc_noserv_addr(address_family,
                             scope,
                             min_desired_addresses,
                             max_desired_addresses,
                             time_now, /* min_desired_start_time */
                             time_now, /* max_desired_start_time */
                             min_desired_lifetime,
                             max_desired_lifetime,
                             multicast_address_set_lease,
                             status);
}

/**
 * This function is used to change the lifetime of an allocation.
 * The actual message sent to the server is a RENEW message.
 *
 * @param multicast_addres_set_lease The lease to renew.
 * @param min_desired_lifetime The minimum requested lifetime.
 * @param max_desired_lifetime The maximum requested lifetime.
 * @param lifetime The new lifetime will be returned here.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 * Possible errors are MC_EFAMILY, MC_ELTIME, MC_EFAILED and MC_ELEASE.
 */
int change_multicast_addr_lifetime(lease *multicast_address_set_lease,
                                   uint32 min_desired_lifetime,
                                   uint32 max_desired_lifetime,
                                   uint32 *lifetime){
    mc_msg              *send_msg      = NULL;
    mc_msg              *reply_msg     = NULL;
    mc_option_lid       *lid_opt       = NULL;
    mc_option_ltime     *ltime_opt     = NULL;
    mc_option_min_ltime *min_ltime_opt = NULL;
    uint8               *raw_message   = NULL;
    int                 message_size   = 0;
    int                 socket         = -1;
    int                 timeout        = -1;
    
    /* First check the arguments. */
    if(multicast_address_set_lease == NULL) {
        mc_errno = MC_ELEASE;
        DEBUG("Lease NULL in change_multicast_addr_lifetime.");
        return -1;
    }
    if(multicast_address_set_lease->address_family != MC_IPV4) {
        mc_errno = MC_EFAMILY;
        DEBUG("Invalid address family in change_multicast_addr_lifetime.");
        return -1;
    }

    if(max_desired_lifetime < min_desired_lifetime) {
        mc_errno = MC_ELTIME;
        DEBUG("Requested lifetime invalid in change_multicast_addr_lifetime.");
        return -1;
    }

    lid_opt = option_create_lid(multicast_address_set_lease->lease_id,
                                multicast_address_set_lease->id_len);
    
    send_msg = msg_create_renew(lid_opt);

    ltime_opt = option_create_ltime(max_desired_lifetime);
    msg_add_option(send_msg, (mc_option *)ltime_opt);
    
    min_ltime_opt = option_create_min_ltime(min_desired_lifetime);
    msg_add_option(send_msg, (mc_option *)min_ltime_opt);

    /* Create and add the End option. */
    msg_add_option(send_msg, (mc_option *)option_create_end());

    /* Encode message. */
    raw_message = mc_encode_message(send_msg, &message_size);

    /* Send message. */
    socket = mc_send_message(raw_message, message_size,
                             multicast_address_set_lease->server_address,
                             MC_SERVERPORT, 0);
    free(raw_message);
    raw_message = NULL;

    /* Receive reply. */
    timeout = MC_BASE_TIMEOUT;
    while((timeout <= 16) && (timeout <= (MC_BASE_TIMEOUT * 4))) {
        raw_message = mc_receive_message(socket, timeout, &message_size);
        if(raw_message == NULL) {
            if(mc_errno == MC_ETIMEOUT) {
                /* Time-out so we increase timeout and try again. */
                timeout = timeout * 2;
                continue;
            } else if(mc_errno == MC_EINTR) {
                /* Interrupted by something else then time-out. Try again. */
                continue;
            }
        }
        break;
    }

    if(raw_message == NULL) {
        DEBUG("Reception of message failed in change_multicast_addr_lifetime.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Decode message. */
    reply_msg = mc_decode_message(raw_message, message_size);
    free(raw_message);
    raw_message = NULL;
    if(reply_msg == NULL) {
        DEBUG("Failed to decode message in change_multicast_addr_lifetime.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Validate message. */
    if(msg_validate(reply_msg, send_msg) == -1) {
        msg_delete(send_msg);
        msg_delete(reply_msg);
        DEBUG("Failed to validate message in change_multicast_addr_lifetime.");
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* No further need for the sent message. */
    msg_delete(send_msg);
    send_msg = NULL;

    if(reply_msg->msgtype == MC_NAK) {
        /* Easy case, we failed. */
#ifdef DEBUGGING
        print_ecode(reply_msg);
#endif
        msg_delete(reply_msg);
        DEBUG("Received NAK in change_multicast_addr_lifetime.");
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Extract the new lifetime. */
    *lifetime =
        ((mc_option_ltime *)msg_get_option(reply_msg, MC_LTIME))->lease_time;
    msg_delete(reply_msg);
    reply_msg = NULL;
    
    return 0;
}

/**
 * This function deallocates a lease allocated earlier.
 *
 * @param multicast_addresset_lease The lease which should be released.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 * Possible errors are MC_EFAMILY, MC_ELEASE and MC_EFAILED.
 */
int deallocate_multicast_addr(lease *multicast_address_set_lease){
    mc_msg              *send_msg      = NULL;
    mc_msg              *reply_msg     = NULL;
    mc_option_lid       *lid_opt       = NULL;

    uint8 *raw_message = NULL;
    int message_size = 0;
    int socket = -1;
    int timeout = -1;

    /* First check the arguments. */
    if(multicast_address_set_lease == NULL) {
        DEBUG("lease == NULL in deallocate_multicast_addr");
        mc_errno = MC_ELEASE;
        return -1;
    }
    if(multicast_address_set_lease->address_family != MC_IPV4) {
        mc_errno = MC_EFAMILY;
        DEBUG("Address family invalid in deallocate_multicast_addr");
        return -1;
    }

    lid_opt = option_create_lid(multicast_address_set_lease->lease_id,
                                multicast_address_set_lease->id_len);
    send_msg = msg_create_release(lid_opt);

    /* Create and add the End option. */
    msg_add_option(send_msg, (mc_option *)option_create_end());

    /* Encode message. */
    raw_message = mc_encode_message(send_msg, &message_size);

    /* Send message. */
    socket = mc_send_message(raw_message,
                             message_size,
                             multicast_address_set_lease->server_address,
                             MC_SERVERPORT, 0);
    free(raw_message);
    raw_message = NULL;

    /* Receive reply. */
    timeout = MC_BASE_TIMEOUT;
    while((timeout <= 16) && (timeout <= (MC_BASE_TIMEOUT * 4))) {
        raw_message = mc_receive_message(socket, timeout, &message_size);
        if(raw_message == NULL) {
            if(mc_errno == MC_ETIMEOUT) {
                /* Time-out so we increase timeout and try again. */
                timeout = timeout * 2;
                continue;
            } else if(mc_errno == MC_EINTR) {
                /* Interrupted by something else then time-out. Try again. */
                continue;
            }
        }
        break;
    }

    if(raw_message == NULL) {
        DEBUG("Reception of message failed in deallocate_multicast_addr.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Decode message. */
    reply_msg = mc_decode_message(raw_message, message_size);
    free(raw_message);
    raw_message = NULL;
    if(reply_msg == NULL) {
        DEBUG("Failed to decode message in deallocate_multicast_addr.");
        msg_delete(send_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* Validate message. */
    if(msg_validate(reply_msg, send_msg) == -1) {
        msg_delete(send_msg);
        msg_delete(reply_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* No further need for the sent message. */
    msg_delete(send_msg);
    send_msg = NULL;

    if(reply_msg->msgtype == MC_NAK) {
        /* Easy case, we failed. */
#ifdef DEBUGGING
        print_ecode(reply_msg);
#endif
        msg_delete(reply_msg);
        mc_errno = MC_EFAILED;
        return -1;
    }

    /* We must have succeeded. */
    msg_delete(reply_msg);
    return 0;
}

int mc_errno = 0;

/* NOTE! No error message may be more than 255 characters + '\0'. */
static char err_string[256];

/* API error strings. */
static char *err_api_errors[] = {
    "Invalid error code = 0.",
    "Bad address family.",
    "Bad scope.",
    "Bad desired number of addresses.",
    "Bad desired lifetimes.",
    "Bad desired start time durations.",
    "Requested start times conflict with requested lifetimes.",
    "Operation failed.",
    "Invalid or expired lease.",
    "Nesting state undetermined at this time."
};

/* Internal error strings. */
static char *err_int_errors[] = {
    "Invalid error code = 100.",
    "Function parameter invalid.",
    "The operation was interrupted.",
    "The operation failed.",
    "The message contained illegal options.",
    "A required option was missing.",
    "Option was duplicated.",
    "The xid of sent and received message did not match.",
    "The reply had an unexpected message type.",
    "Lease identifiers did not match.",
    "Timeout in receive.",
    "The message was too large, or too little data was sent.",
    "The message could not be decoded.",
    "The message used the wrong version of the MADCAP protocol."
};

/**
 * Get a string representation of the error.
 *
 * @param errnum The error code.
 * @return A pointer to a string representing the error. The string may only be
 * used until the next call to this function.
 */
char *mc_strerror(int errnum){
    if(errnum > MC_MIN_API_ERRNO && errnum < MC_MAX_API_ERRNO) {
        strcpy(err_string, err_api_errors[errnum]);
    } else if(errnum > MC_MIN_INT_ERRNO && errnum < MC_MAX_INT_ERRNO) {
        strcpy(err_string, err_int_errors[errnum - 100]);
    } else {
        strcpy(err_string, "Invalid error code.");
    }
    return err_string;
}

/**
 * Print a description of the error to the standard error output.
 *
 * @param s A string which will be printed first. After the string there will
 * be a ':', a space and then the error message of the latest error followed by
 * a newline.
 */
void mc_perror(const char *s){
    fprintf(stderr, "%s: %s\n", s, mc_strerror(mc_errno));
}
