/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the definition of the libmadcap API.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _madcap_h_
#define _madcap_h_

#include <stdio.h> /* For the DEBUG macro. */
#include <stdlib.h> /* For the ABORT macro. */

/* Types used for exact sizes of variables. */
typedef signed char    int8;
typedef unsigned char  uint8;
typedef signed short   int16;
typedef unsigned short uint16;
typedef signed long    int32;
typedef unsigned long  uint32;

/* Constants. */
enum{
    MC_SOK          = 1,     /* Status OK.                                    */
    MC_STRY_LATER   = 2,     /* Status try later.                             */

    MC_IPV4         = 1,     /* IPv4 address family.                          */
    MC_IPV6         = 2,     /* IPv6 address family.                          */

    MC_VERSION      = 0,     /* MADCAP protocol version.                      */

    MC_RECVBUFSIZE  = 10240, /* The size of the buffer in mc_receive_message. */

    MC_SERVERPORT   = 2535,  /* The port where the MAAS is running.           */

    MC_BASE_TIMEOUT = 4      /* Time-out in seconds.                          */
};

/* An address range. The range is [start, end]. */
typedef struct address_range{
    uint32 start;
    uint32 end;
} address_range;

/* Represents a name, for use in the scopes. */
typedef struct name{
    char *language_tag; /* NULL terminated string identifying the language used
                           for this name. */
    char *name;         /* NULL terminated string containing the name. */
} name;

/* Represents a multicast scope. */
typedef struct scope{
    address_range addresses; /* The set of addresses in the scope. */
    name *name_set;          /* The names for the scope. */
    unsigned int num_names;  /* The number of names for the scope. */
    uint8 ttl;               /* Default TTL. */
} scope;

/* Represents a lease to the client. */
typedef struct lease{
    uint32 *addresses;     /* The set of addresses leased. */
    unsigned int num_addr; /* The number of addresses leased. */
    uint16 address_family; /* The address family of the lease. */
    uint32 lifetime;       /* How long are the addresses valid? */
    uint32 start_time;     /* When the lease start beeing valid. */
    uint8  *lease_id;      /* Unique identifier of this lease. */
    unsigned int id_len;   /* The length of the lease id. */
    uint32 server_address; /* The server which gave us the lease. */
} lease;

/* API functions. */

/**
 * General function for allocation of multicast addresses.
 * This function is also called by the more specialized allocation functions.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param server The server from which to allocate the address.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
extern int alloc_multicast_addr(uint16 address_family,
                                scope *scope,
                                uint16 min_desired_addresses,
                                uint16 max_desired_addresses,
                                uint32 min_desired_start_time,
                                uint32 max_desired_start_time,
                                uint32 min_desired_lifetime,
                                uint32 max_desired_lifetime,
                                uint32 server,
                                lease *multicast_address_set_lease,
                                int *status);

/**
 * General function for allocation of multicast addresses when the server is not
 * known.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
extern int alloc_noserv_addr(uint16 address_family,
                             scope *scope,
                             uint16 min_desired_addresses,
                             uint16 max_desired_addresses,
                             uint32 min_desired_start_time,
                             uint32 max_desired_start_time,
                             uint32 min_desired_lifetime,
                             uint32 max_desired_lifetime,
                             lease *multicast_address_set_lease,
                             int *status);

/**
 * Allocate a single multicast address.
 * This function calls the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
extern int alloc_one_addr(uint16 address_family,
                          scope *scope,
                          uint32 min_desired_start_time,
                          uint32 max_desired_start_time,
                          uint32 min_desired_lifetime,
                          uint32 max_desired_lifetime,
                          lease *multicast_address_set_lease,
                          int *status);

/**
 * Function for allocation of addresses with an exact lifetime.
 * This function will call the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_start_time The earliest time at which the allocation
 * should start.
 * @param max_desired_start_time The latest time at which the allocation
 * should start.
 * @param lifetime The time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
extern int alloc_exact_addr(uint16 address_family,
                            scope *scope,
                            unsigned int min_desired_addresses,
                            unsigned int max_desired_addresses,
                            uint32 min_desired_start_time,
                            uint32 max_desired_start_time,
                            uint32 lifetime,
                            lease *multicast_address_set_lease,
                            int *status);

/**
 * Function for allocation of addresses which starts immediately.
 * This function will call the general alloc_multicast_addr function.
 *
 * @param address_family The address family where the allocation should be done.
 * Can currently only be MC_IPV4.
 * @param scope The scope in which to allocate the addresses.
 * @param min_desired_addresses The minimum number of addresses we wish to
 * allocate.
 * @param max_desired_addresses The maximum number of addresses we wish to
 * allocate.
 * @param min_desired_lifetime The earliest time at which the allocation may
 * end.
 *@param max_desired_lifetime The latest time at which the allocation may end.
 * @param multicast_address_set_lease The lease identifying a successful
 * allocation is returned in this struct. It does not contain any valid data if
 * the allocation failed.
 * @param status The return status of the allocation. May be MC_SOK or
 * MC_STRY_LATER. The value of status is not defined when the function failed.
 * @return A return value of 0 indicates a success and -1 indicates failure. The
 * value of the status parameter must also be checked if the return value is 0.
 * Possible errors are MC_EFAMILY, MC_ESCOPE, MC_ENUMADDR, MC_ELTIME, MC_ESTIME,
 * MC_ETIMECONF and MC_EFAILED.
 */
extern int alloc_immediate_addr(uint16 address_family,
                                scope *scope,
                                int min_desired_addresses,
                                int max_desired_addresses,
                                uint32 min_desired_lifetime,
                                uint32 max_desired_lifetime,
                                lease *multicast_address_set_lease,
                                int *status);

/**
 * This function is used to change the lifetime of an allocation.
 * The actual message sent to the server is a RENEW message.
 *
 * @param multicast_addres_set_lease The lease to renew.
 * @param min_desired_lifetime The minimum requested lifetime.
 * @param max_desired_lifetime The maximum requested lifetime.
 * @param lifetime The new lifetime will be returned here.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 * Possible errors are MC_EFAMILY, MC_ELTIME, MC_EFAILED and MC_ELEASE.
 */
extern int change_multicast_addr_lifetime(lease *multicast_address_set_lease,
                                          uint32 min_desired_lifetime,
                                          uint32 max_desired_lifetime,
                                          uint32 *lifetime);

/**
 * This function deallocates a lease allocated earlier.
 *
 * @param multicast_addresset_lease The lease which should be released.
 * @return A return value of 0 indicates a success and -1 indicates failure.
 * Possible errors are MC_EFAMILY and MC_ELEASE.
 */
extern int deallocate_multicast_addr(lease *multicast_address_set_lease);

/**
 * This function is NOT implemented.
 */
extern int get_multicast_addr_scope(uint16 address_family,
                                    scope *scope_set,
                                    unsigned int *num_scopes);

/**
 * This function is NOT implemented.
 */
extern int get_scope_name(scope scope,
                          char *language_tag,
                          char *name,
                          char *language_for_name);

/**
 * This function is NOT implemented.
 * NOTE: The nesting_relationship parameter may be of wrong type!
 */
extern int get_scope_nesting_state(scope *scope_set,
                                   unsigned int num_scopes,
                                   int **nesting_relationship);

/**
 * This function is NOT implemented.
 */
extern int get_larger_scopes(scope s,
                             scope *scope_set,
                             unsigned int *num_scopes);

/**
 * This function is NOT implemented.
 */
extern int get_smaller_scopes(scope s,
                              scope *scope_set,
                              unsigned int *num_scopes);

/**
 * This function is NOT implemented.
 */
extern int change_multicast_addr_start_time(lease multicast_addresset_lease,
                                            uint32 min_desired_start_time,
                                            uint32 max_desired_start_time,
                                            uint32 *start_time);


/* Error related stuff. */

/* The set of possible error codes. */
enum{
    /* API Errors. */
    MC_MIN_API_ERRNO = 0,
    MC_EFAMILY       = 1,
    MC_ESCOPE        = 2,
    MC_ENUMADDR      = 3,
    MC_ELTIME        = 4,
    MC_ESTIME        = 5,
    MC_ETIMECONF     = 6,
    MC_EFAILED       = 7,
    MC_ELEASE        = 8,
    MC_ENONEST       = 9,
    MC_MAX_API_ERRNO = 10,

    /* Internal errors. */
    MC_MIN_INT_ERRNO = 100,
    MC_EPARAM        = 101, /* any: function parameter was invalid. */
    MC_EINTR         = 102, /* any: operation was interrupted. */
    MC_EFAIL         = 103, /* any: operation failed. */
    MC_EOPTION       = 104, /* validate: invalid option for the message. */
    MC_ENOOPTION     = 105, /* validate: required option was missing. */
    MC_EDUPOPTION    = 106, /* validate: option was duplicated. */
    MC_EXID          = 107, /* validate: xid of messages did not match. */
    MC_EREPLY        = 108, /* validate: unexpected reply type. */
    MC_ELID          = 109, /* validate: lease identifier mismatch. */
    MC_ETIMEOUT      = 110, /* receive: time-out occurred. */
    MC_ESIZE         = 111, /* receive: message was too large. send: too little
                             data was sent. */
    MC_EBROKEN       = 112, /* decode: the raw message could not be decoded. */
    MC_EVERSION      = 113, /* decode: Wrong version of the MADCAP protocol. */
    MC_MAX_INT_ERRNO = 114
};

/** This variable should contain the error code after an error has occured. */
extern int mc_errno;

/**
 * Get a string representation of the error.
 *
 * @param errnum The error code.
 * @return A pointer to a string representing the error. The string may only be
 * used until the next call to this function.
 */
extern char *mc_strerror(int errnum);

/**
 * Print a description of the error to the standard error output.
 *
 * @param s A string which will be printed first. After the string there will
 * be a ':', a space and then the error message of the latest error followed by
 * a newline.
 */
extern void mc_perror(const char *s);

/**
 * Use this instead of exit(EXIT_FAILURE).
 */
#define ABORT(s) do{                                                           \
    fprintf(stderr, "Execution aborted at %s[%d]: %s\n",                       \
            __FILE__, __LINE__, s);                                            \
    exit(EXIT_FAILURE);                                                        \
}while(0)

/**
 * Used to test conditions.
 */
#define ASSERT(c, s) do{                                                       \
    if (!(c)) {                                                                \
        fprintf(stderr, "Assertion failed at %s[%d]. %s\n",                    \
                __FILE__, __LINE__, s);                                        \
        exit(EXIT_FAILURE);                                                    \
    }                                                                          \
}while(0)

#endif
