/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the message handling code.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#include "message.h"
#include "list.h"
#include "net.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>

/**
 * Used to create a REQUEST message.
 *
 * @param sid The server identifier option.
 * @param lid The lease identifier option.
 * @param scope The scope option.
 * @return The newly created message. It should be deleted using msg_delete.
 * A return value of NULL indicates a failure to allocate memory for the new
 * message.
 */
mc_msg *msg_create_request(mc_option_sid *sid,
                           mc_option_lid *lid,
                           mc_option_scope *scope){
    mc_msg *ret_msg = (mc_msg *)malloc(sizeof(mc_msg));
    ASSERT(ret_msg != NULL, "Memory allocation failed.");

    ret_msg->version    = 0;
    ret_msg->msgtype    = MC_REQUEST;
    ret_msg->addrfamily = MC_IPV4;
    ret_msg->xid        = mc_get_xid();

    ret_msg->options    = list_create();
    ASSERT(ret_msg->options != NULL, "List creation failed.");
    msg_add_option(ret_msg, (mc_option *)sid);
    msg_add_option(ret_msg, (mc_option *)lid);
    msg_add_option(ret_msg, (mc_option *)scope);

    return ret_msg;
}

/**
 * Used to create a RENEW message.
 *
 * @param lid The lease identifier option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
mc_msg *msg_create_renew(mc_option_lid *lid){
    mc_msg *ret_msg = (mc_msg *)malloc(sizeof(mc_msg));
    ASSERT(ret_msg != NULL, "Memory allocation failed.");

    ret_msg->version    = 0;
    ret_msg->msgtype    = MC_RENEW;
    ret_msg->addrfamily = MC_IPV4;
    ret_msg->xid        = mc_get_xid();

    ret_msg->options    = list_create();
    ASSERT(ret_msg->options != NULL, "List creation failed.");
    msg_add_option(ret_msg, (mc_option *)lid);

    return ret_msg;
}

/**
 * Used to create a RELEASE message.
 *
 * @param lid The lease identifier option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
mc_msg *msg_create_release(mc_option_lid *lid){
    mc_msg *ret_msg = (mc_msg *)malloc(sizeof(mc_msg));
    ASSERT(ret_msg != NULL, "Memory allocation failed.");

    ret_msg->version    = 0;
    ret_msg->msgtype    = MC_RELEASE;
    ret_msg->addrfamily = MC_IPV4;
    ret_msg->xid        = mc_get_xid();

    ret_msg->options    = list_create();
    ASSERT(ret_msg->options != NULL, "List creation failed.");
    msg_add_option(ret_msg, (mc_option *)lid);

    return ret_msg;
}

/**
 * Used to create a GETINFO message.
 *
 * @param lid The lease identifier option.
 * @param reqlist The option request list option.
 * @return The newly created message. It should be deleted using msg_delete.
 */
extern mc_msg *msg_create_getinfo(mc_option_lid *lid,
                                  mc_option_reqlist *reqlist){
    mc_msg *ret_msg = (mc_msg *)malloc(sizeof(mc_msg));
    ASSERT(ret_msg != NULL, "Memory allocation failed.");

    ret_msg->version    = 0;
    ret_msg->msgtype    = MC_GETINFO;
    ret_msg->addrfamily = MC_IPV4;
    ret_msg->xid        = mc_get_xid();

    ret_msg->options    = list_create();
    ASSERT(ret_msg->options != NULL, "List creation failed.");
    msg_add_option(ret_msg, (mc_option *)lid);
    msg_add_option(ret_msg, (mc_option *)reqlist);

    return ret_msg;
}

/**
 * This function is used to delete a message from memory. It will recursively
 * free the memory used by each option, using option_delete(), and then free
 * the message itself.
 *
 * @param message The message to clear from memory.
 */
void msg_delete(mc_msg *message){
    mc_option *opt = NULL;
    ASSERT(message != NULL, "Message is NULL in msg_delete.");
    opt = list_first(message->options);
    while(opt != NULL) {
        option_delete(opt);
        opt = list_next(message->options);
    }
    list_destroy(message->options);
    free(message);
}

/**
 * This is used to add an option to the list of options in a message.
 *
 * @param msg The message where the option should be added.
 * @param option The option which should be added.
 */
void msg_add_option(mc_msg *msg, mc_option *option){
    list_last(msg->options);
    list_insert(msg->options, option, LIST_AFTER);
}

/**
 * This is used to get a specific option from a message.
 *
 * @param msg The message to work with.
 * @param option The code of the option to retreive.
 * @return A pointer to the option, or NULL if the option did not exist.
 */
mc_option *msg_get_option(mc_msg *msg, uint16 option){
    mc_option *opt = (mc_option *)list_first(msg->options);
    while(opt != NULL) {
        if(opt->code == option) {
            break;
        }
        opt = (mc_option *)list_next(msg->options);
    }
    return opt;
}

/**
 * This function is used to check the syntax of a reply message received from a
 * server.
 *
 * @param message The message to check.
 * @param original The message which resulted in the reply.
 * @return 0 if the message was ok, -1 if there was an error. Possible error
 * values are MC_EPARAM, MC_EOPTION, MC_ENOOPTION, MC_EDUPOPTION and MC_EREPLY.
 */
int msg_validate(mc_msg *message, mc_msg *original){
    mc_option *opt1 = NULL;
    mc_option *opt2 = NULL;
    int index1 = 0, index2 = 0;
    
    if((message == NULL) || (original == NULL)) {
        DEBUG("(message == NULL) || (original == NULL) in msg_validate.");
        mc_errno = MC_EPARAM;
        return -1;
    }

    /* Check that the xid is the same. */
    if(message->xid != original->xid) {
        DEBUG("Different xid in msg_validate.");
        mc_errno = MC_EXID;
        return -1;
    }

    /* Check that there are options in the message. */
    if(list_is_empty(message->options) == 1) {
        DEBUG("No options in the message in msg_validate.");
        mc_errno = MC_ENOOPTION;
        return -1;
    }

    /* Check that no option is there twice. */
    opt1 = (mc_option *)list_jump(message->options, index1);
    while(opt1 != NULL) {
        index2 = index1 + 1;
        opt2 = (mc_option *)list_jump(message->options, index2);
        while(opt2 != NULL) {
            if(opt1->code == opt2->code) {
                DEBUG("Duplicate options in message in msg_validate");
                mc_errno = MC_EDUPOPTION;
                return -1;
            }
            index2++;
            opt2 = (mc_option *)list_jump(message->options, index2);
        }
        index1++;
        opt1 = (mc_option *)list_jump(message->options, index1);
    }

    /* Check that the last option is the End option. */
    opt1 = (mc_option *)list_last(message->options);
    if(opt1->code != MC_END) {
        DEBUG("No End option in the message in msg_validate.");
        mc_errno = MC_ENOOPTION;
        return -1;
    }

    /* Now for the message type specific checks. */
    switch(original->msgtype) {
    case MC_REQUEST:
        switch(message->msgtype) {
        case MC_ACK:
            if(!msg_get_option(message, MC_LTIME) ||
               !msg_get_option(message, MC_SID)   ||
               !msg_get_option(message, MC_LID)   ||
               !msg_get_option(message, MC_SCOPE) ||
               !msg_get_option(message, MC_ADDRLIST)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME) ||
               msg_get_option(message, MC_ERROR)) {
                DEBUG("ACK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_STIME) &&
               !msg_get_option(message, MC_CTIME)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_RTIME) &&
               !msg_get_option(message, MC_CTIME)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))
                       ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        case MC_NAK:
            if(!msg_get_option(message, MC_SID)   ||
               !msg_get_option(message, MC_LID)   ||
               !msg_get_option(message, MC_ERROR)) {
                DEBUG("NAK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_LTIME)     ||
               msg_get_option(message, MC_SCOPE)     ||
               msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_STIME)     ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_ADDRLIST)  ||
               msg_get_option(message, MC_CTIME)     ||
               msg_get_option(message, MC_RTIME)     ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME)) {
                DEBUG("NAK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))                               ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        default:
            DEBUG("Unexpected reply to Request message.");
            mc_errno = MC_EREPLY;
            return -1;
            break;
        }
        break;
    case MC_RELEASE:
        switch(message->msgtype) {
        case MC_ACK:
            if(!msg_get_option(message, MC_SID) ||
               !msg_get_option(message, MC_LID)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_LTIME)     ||
               msg_get_option(message, MC_SCOPE)     ||
               msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_STIME)     ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_ADDRLIST)  ||
               msg_get_option(message, MC_CTIME)     ||
               msg_get_option(message, MC_RTIME)     ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME) ||
               msg_get_option(message, MC_ERROR)) {
                DEBUG("ACK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))
                       ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        case MC_NAK:
            if(!msg_get_option(message, MC_SID)   ||
               !msg_get_option(message, MC_LID)   ||
               !msg_get_option(message, MC_ERROR)) {
                DEBUG("NAK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_LTIME)     ||
               msg_get_option(message, MC_SCOPE)     ||
               msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_STIME)     ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_ADDRLIST)  ||
               msg_get_option(message, MC_CTIME)     ||
               msg_get_option(message, MC_RTIME)     ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME)) {
                DEBUG("NAK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))
                       ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        default:
            DEBUG("Unexpected reply to Release message.");
            mc_errno = MC_EREPLY;
            return -1;
            break;
        }
        break;
    case MC_GETINFO:
        switch(message->msgtype) {
        case MC_ACK:
            if(!msg_get_option(message, MC_SID) ||
               !msg_get_option(message, MC_LID)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_LTIME)     ||
               msg_get_option(message, MC_SCOPE)     ||
               msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_STIME)     ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_ADDRLIST)  ||
               msg_get_option(message, MC_RTIME)     ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME) ||
               msg_get_option(message, MC_ERROR)) {
                DEBUG("ACK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))
                       ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        default:
            DEBUG("Unexpected reply to Getinfo message.");
            mc_errno = MC_EREPLY;
            return -1;
            break;
        }
        break;
    case MC_RENEW:
        switch(message->msgtype) {
        case MC_ACK:
            if(!msg_get_option(message, MC_LTIME) ||
               !msg_get_option(message, MC_SID)   ||
               !msg_get_option(message, MC_LID)   ||
               !msg_get_option(message, MC_SCOPE) ||
               !msg_get_option(message, MC_ADDRLIST)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME) ||
               msg_get_option(message, MC_ERROR)) {
                DEBUG("ACK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_STIME) &&
               !msg_get_option(message, MC_CTIME)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_RTIME) &&
               !msg_get_option(message, MC_CTIME)) {
                DEBUG("ACK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))
                       ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        case MC_NAK:
            if(!msg_get_option(message, MC_SID)   ||
               !msg_get_option(message, MC_LID)   ||
               !msg_get_option(message, MC_ERROR)) {
                DEBUG("NAK message was missing a MUST option in msg_validate.");
                mc_errno = MC_ENOOPTION;
                return -1;
            }
            if(msg_get_option(message, MC_LTIME)     ||
               msg_get_option(message, MC_SCOPE)     ||
               msg_get_option(message, MC_REQLIST)   ||
               msg_get_option(message, MC_STIME)     ||
               msg_get_option(message, MC_NUMADDR)   ||
               msg_get_option(message, MC_REQLANG)   ||
               msg_get_option(message, MC_SCOPELIST) ||
               msg_get_option(message, MC_ADDRLIST)  ||
               msg_get_option(message, MC_CTIME)     ||
               msg_get_option(message, MC_RTIME)     ||
               msg_get_option(message, MC_MIN_LTIME) ||
               msg_get_option(message, MC_MAX_STIME)) {
                DEBUG("NAK message contained MUST NOT option in msg_validate.");
                mc_errno = MC_EOPTION;
                return -1;
            }
            if((((mc_option_lid *)msg_get_option(message, MC_LID))->len !=
                ((mc_option_lid *)msg_get_option(original, MC_LID))->len) ||
               (memcmp(((mc_option_lid *)msg_get_option(message, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(original, MC_LID))->id,
                       ((mc_option_lid *)msg_get_option(message, MC_LID))                               ->len))) {
                DEBUG("Differing lease identifiers in msg_validate.");
                mc_errno = MC_ELID;
                return -1;
            }
            break;
        default:
            DEBUG("Unexpected reply to Renew message.");
            mc_errno = MC_EREPLY;
            return -1;
            break;
        }
        break;
    default:
        ABORT("Validation of that message type not implemented yet.");
        break;
    }
    
    return 0;
}
