/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the network handling code. For the implementation of the
 * functions defined here, please see net.c.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _net_h_
#define _net_h_

#include "madcap.h" /* For the data types. */
#include "message.h"
#include "option.h"

/**
 * This function is used to encode a message. The message will be copied to a
 * block of memory and will there have the same representation as it should have
 * when sent over the network.
 *
 * @param message The message to encode.
 * @param message_size The size if the encoded message will be stored here.
 * @return A block of memory containing the message in network format.
 */
extern uint8 *mc_encode_message(mc_msg *message, int *message_size);

/**
 * This function is used to encode an option and add it to the end of an
 * existing, already encoded, message.
 *
 * @param option The option to encode.
 * @param raw_message The message to add to option to.
 * @param index Where in the message to add the option. The index will be
 * updated as a result of calling this function.
 * @return The newly encoded message. The pointer may be the same as the
 * parameter message, but may also have changed.
 */
extern uint8 *mc_encode_option(mc_option *option,
                               uint8 *raw_message,
                               int *index);

/**
 * This function will send an encoded message to a MAAS server or to a multicast
 * address.
 *
 * @param data A pointer to the encoded data.
 * @param size The number of bytes in the message.
 * @param address The address to the server or multicast group. It should be in
 * host byte order.
 * @param port The port the message will be sent over. It should be in host byte
 * order.
 * @param ttl The time-to-live which should be used for this message if it is
 * sent over a multicast address.
 * @return The socket where the message was sent, or -1 if there was a problem.
 */
extern int mc_send_message(uint8 *data,
                           int size,
                           uint32 address,
                           uint16 port,
                           uint8 ttl);

/**
 * This function will receive a message from the network.
 *
 * @param socket The socket where the message should be received. The socket
 * will not be closed after a successful reception.
 * @param timeout The period of time to wait before failing with an error. Time
 * is in milliseconds.
 * @param message_size The size of the received message will be stored here.
 * @return A pointer to the raw data received from the network. NULL if the
 * reception failed. Possible errors are MC_EPARAM, MC_EINTR, MC_EFAIL,
 * MC_ETIMEOUT and MC_ESIZE.
 */
extern uint8 *mc_receive_message(int socket,
                                 unsigned int timeout,
                                 int *message_size);

/**
 * This function will decode a message as it comes from the receive function and
 * create a mc_msg to store it in. No validation is done here.
 *
 * @param raw_message The raw message.
 * @param size The number of bytes in the message.
 * @return A pointer to the decoded message, or NULL if decoding failed.
 */
extern mc_msg *mc_decode_message(uint8 *raw_message, int size);

/**
 * This function will decode an option as it comes from the receive function and
 * create a mc_option to store it in. It is used by mc_decode_message.
 *
 * @param raw_message The raw message.
 * @param size The number of bytes in the message.
 * @return A pointer to the decoded message, or NULL if decoding failed.
 */
extern mc_option *mc_decode_option(uint8 *raw_message, int size, int *index);

/**
 * Get a new transaction identifier.
 *
 * @return The new trasaction identifier.
 */
extern uint32 mc_get_xid();

#endif
