/*
 * This is the madcap library, an implementation of the MADCAP protocol for
 * multicast address allocation.
 *
 * This file contains the option handling code. For the implementation of the
 * functions defined here, please see option.c.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _option_h_
#define _option_h_

#include "madcap.h" /* For the data types. */
#include "list.h"

/* These are the different options a MADCAP message may carry. */
enum{
    MC_END         = 0,
    MC_LTIME       = 1,
    MC_SID         = 2,
    MC_LID         = 3,
    MC_SCOPE       = 4,
    MC_REQLIST     = 5,
    MC_STIME       = 6,
    MC_NUMADDR     = 7,
    MC_REQLANG     = 8,
    MC_SCOPELIST   = 9,
    MC_ADDRLIST    = 10,
    MC_CTIME       = 11,
    MC_FEATURELIST = 12,
    MC_RTIME       = 13,
    MC_MIN_LTIME   = 14,
    MC_MAX_STIME   = 15,
    MC_ERROR       = 16
};

/* This is the basic look of an option. */
typedef struct mc_option{
    uint16 code;
} mc_option;

/* The end option. */
typedef struct mc_option_end{
    uint16 code;
    uint16 len;
} mc_option_end;

/* The lease time option. */
typedef struct mc_option_ltime{
    uint16 code;
    uint16 len;
    uint32 lease_time;
} mc_option_ltime;

/* The server identifier option. */
typedef struct mc_option_sid{
    uint16 code;
    uint16 len;
    uint16 addr_family;
    uint32 address;
} mc_option_sid;

/* The lease identifier option. */
typedef struct mc_option_lid{
    uint16 code;
    uint16 len;
    uint8  *id;
} mc_option_lid;

/* The scope option. */
typedef struct mc_option_scope{
    uint16 code;
    uint16 len;
    uint32 scope_id;
} mc_option_scope;

/* The Option Request List option. */
typedef struct mc_option_reqlist{
    uint16 code;
    uint16 len;
    list *options; /* List contains uint16. */
} mc_option_reqlist;

/* The Start Time option. */
typedef struct mc_option_stime{
    uint16 code;
    uint16 len;
    uint32 start_time;
} mc_option_stime;

/* The Number of Addresses Requested option. */
typedef struct mc_option_numaddr{
    uint16 code;
    uint16 len;
    uint16 min;
    uint16 desired;
} mc_option_numaddr;

/* The Requested Language option. */
typedef struct mc_option_reqlang{
    uint16 code;
    uint16 len;
    uint8  *tag;
} mc_option_reqlang;

/* The Multicast Scope List option. */
typedef struct mc_option_scopelist{
    uint16 code;
    uint16 len;
    list *scopes; /* List contains scope. */
} mc_option_scopelist;

/* The List of Address Ranges option. */
typedef struct mc_option_addrlist{
    uint16 code;
    uint16 len;
    list *ranges; /* List contains address_range. */
} mc_option_addrlist;

/* The Current Time option. */
typedef struct mc_option_ctime{
    uint16 code;
    uint16 len;
    uint32 time;
} mc_option_ctime;

/* The Feature List option. */
typedef struct mc_option_featurelist{
    uint16 code;
    list *supported; /* List contains uint16. */
    list *requested; /* List contains uint16. */
    list *required;  /* List contains uint16. */
} mc_option_featurelist;

/* The Retry Time option. */
typedef struct mc_option_rtime{
    uint16 code;
    uint16 len;
    uint32 retry_time;
} mc_option_rtime;

/* The Minimum Lease Time option. */
typedef struct mc_option_min_ltime{
    uint16 code;
    uint16 len;
    uint32 min_ltime;
} mc_option_min_ltime;

/* The Maximum Start Time option. */
typedef struct mc_option_max_stime{
    uint16 code;
    uint16 len;
    uint32 max_stime;
} mc_option_max_stime;

/* The Error option. */
typedef struct mc_option_error{
    uint16 code;
    uint16 len;
    uint16 ecode;
    uint8  *data;
} mc_option_error;

/**
 * Used to create an End option.
 *
 * @return An End option.
 */
extern mc_option_end *option_create_end();

/**
 * Used to create a Lease Time option.
 *
 * @param lease_time The lease time in units of seconds.
 * @return A Lease Time option.
 */
extern mc_option_ltime *option_create_ltime(uint32 lease_time);

/**
 * Used to create a Server ID option.
 *
 * @param address The address of the server.
 * @return A Server ID option.
 */
extern mc_option_sid *option_create_sid(uint32 address);

/**
 * Used to create a Lease ID option.
 *
 * @param id The id to use in the option. If it is NULL then a 16 byte id will
 * be created.
 * @param len The length of the id option.
 * @return A Lease ID option.
 */
extern mc_option_lid *option_create_lid(uint8 *id, unsigned int len);

/**
 * Used to create a Scope option.
 *
 * @param scope The first address in the scope.
 * @return A Scope option.
 */
extern mc_option_scope *option_create_scope(uint32 scope);

/**
 * Used to create an Option Request List option.
 *
 * @param len The number of options in the list.
 * @param options The list of requested options.
 * @return An Option Request List option.
 */
extern mc_option_reqlist *option_create_reqlist(uint16 len, uint16 *options);

/**
 * Used to create a Start Time option.
 *
 * @param start_time The start time.
 * @return A Start Time option.
 */
extern mc_option_stime *option_create_stime(uint32 start_time);

/**
 * Used to create a Number of Addresses Requested option.
 *
 * @param min The minimum number of addresses.
 * @param desired The desired number of addresses.
 * @return A Number of Addresses Requested option.
 */
extern mc_option_numaddr *option_create_numaddr(uint16 min, uint16 desired);

/**
 * Used to create a Requested Language option.
 *
 * @param len The number of language tags.
 * @param tag The list of language tags.
 * @return A Requested Language option.
 */
extern mc_option_reqlang *option_create_reqlang(uint16 len, uint8 *tag);

/**
 * Used to create a Multicast Scope List option.
 * @param len The total length of the data of the option when encoded.
 * @param scopes The list of scopes to store in the option.
 */
extern mc_option_scopelist *option_create_scopelist(uint16 len, list *scopes);

/**
 * Used to create a List of Address Ranges option.
 *
 * @param len The number of address ranges.
 * @param ranges The list of address ranges.
 * @return A List of Address Ranges option.
 */
extern mc_option_addrlist *option_create_addrlist(uint16 len, list *ranges);

/**
 * Used to create a Current Time option.
 *
 * @param ctime The current time. If it is 0, then the
 * current time as seen by the local clock will be used.
 * @return A Current Time option.
 */
extern mc_option_ctime *option_create_ctime(uint32 ctime);

/**
 * Used to create a Feature List option.
 *
 * @param supported The list of supported options.
 * @param requested The list of requested options.
 * @param required The list of required options.
 * @return A Feature List option.
 */
extern mc_option_featurelist *option_create_featurelist(list *supported,
                                                        list *requested,
                                                        list *required);

/**
 * Used to create a Retry Time option.
 *
 * @param retry_time The retry time.
 * @return A Retry Time option.
 */
extern mc_option_rtime *option_create_rtime(uint32 retry_time);

/**
 * Used to create a Minimum Lease Time option.
 *
 * @param min_ltime The minimum lease time.
 * @return A Minimum Lease Time option.
 */
extern mc_option_min_ltime *option_create_min_ltime(uint32 min_ltime);

/**
 * Used to create a Maximum Start Time option.
 *
 * @param max_stime The maximum start time.
 * @return A Maximum Start Time option.
 */
extern mc_option_max_stime *option_create_max_stime(uint32 max_stime);

/**
 * Used to create an Error option.
 *
 * @param ecode The error code.
 * @param data The data of the error.
 * @param data_len The number of bytes of data.
 * @return An Error option.
 */
extern mc_option_error *option_create_error(uint16 ecode,
                                            uint8 *data,
                                            unsigned int data_len);

/**
 * This function is used to delete an option from memory.
 *
 * @param option The option to clear from memory.
 */
extern void option_delete(mc_option *option);

#endif
