/**
   The dll module represents a double linked list.
   @author Hkan Bystrm, Linus Svensson, ke stmark
   @version 2000-05-15
*/

#include <stdlib.h>
#include "dll.h"
/*  #include "stdio.h" */

typedef enum place {
    AFTER, BEFORE
} place;

typedef enum status {
    FAILURE=-1,SUCCESS=0
} status;

typedef enum boolean {
    FALSE, TRUE
} boolean;

typedef struct NODE {
    struct NODE *previous;
    struct NODE *next;
    ELEMENT data;
} NODE;

struct DLL {
    NODE *first;
    NODE *current;
    NODE *last;
};

void quicksort(DLL *, NODE *, NODE *, int (*)(ELEMENT, ELEMENT));
NODE *partition(DLL *, NODE *, NODE *, int (*)(ELEMENT, ELEMENT));

/**
   Creates an empty DLL.
   @return a pointer to an ADT on success, NULL on failure
*/
DLL *createDLL(){
    DLL *new = (DLL *) malloc(sizeof(DLL));
    new->first = new->last = new->current = NULL;
    return new;
}

/**
   Removes the ADT from memory.
   @return void.
*/
void destroyDLL(DLL *dll){
    firstElement(dll);
    while (!isEmptyDLL(dll)){
	deleteElement(dll);
    }
    free(dll);
}

/**
   Checks to see if the list is empty.
   @return 1 if list is empty, 0 otherwise
*/
int isEmptyDLL(DLL *dll){
    int result;
    if (dll->current == NULL)
	result = TRUE;
    else
	result = FALSE;
    return result;
}

/**
   Sets "current" element to the next element.
   @return 0 on success, -1 if such an element does not exist.
*/
int nextElement(DLL *dll){
    int result;
    if (isEmptyDLL(dll)) {
	result = FAILURE;
    } else if (dll->current == dll->last) {
	result = FAILURE;
    } else {
	dll->current = dll->current->next;
	result = SUCCESS;
    }
    return result;
}

/**
   Sets "current" element to the prev element.
   @return 0 on success, -1 if such an element does not exist.
*/
int prevElement(DLL *dll){
    int result;
    if (isEmptyDLL(dll)) {
	result = FAILURE;
    } else if (dll->current == dll->first) {
	result = FAILURE;
    } else {
	dll->current = dll->current->previous;
	result = SUCCESS;
    }
    return result;    
}

/**
   Sets "current" element to the first element.
   @return 0 on success, -1 if such an element does not exist.
*/
int firstElement(DLL *dll){
    int result;
    if (isEmptyDLL(dll)) {
	result = FAILURE;
    } else {
	dll->current = dll->first;
	result = SUCCESS;
    }
    return result;
}

/**
   Sets "current" element to the last element.
   @return 0 on success, -1 if such an element does not exist.
*/
int lastElement(DLL *dll){
    int result;
    if (isEmptyDLL(dll)) {
	result = FAILURE;
    } else {
	dll->current = dll->last;
	result = SUCCESS;
    }
    return result;
}

/**
   Gets the current element.
   @return "current" element, or NULL if such an element does not exist.
*/
ELEMENT thisElement(DLL *dll){
    ELEMENT result = NULL;
    if (!isEmptyDLL(dll))
	result = (dll->current->data);
    return result;
}

/**
   Insert an element in the DLL.
   @precondition Inserts an element in the list. If where == 0 the element is inserted after "current" element, otherwise it is inserted before the current element.
   @postcondition The current element must be set to the inserted element.
   @return 0 on success, -1 on failure.
*/
int insertElement(DLL *dll, ELEMENT element, int where){
    NODE *new = (NODE *) malloc(sizeof(NODE));
    int result;
    new->data = element;
    if (isEmptyDLL(dll)) {
	new->next = NULL;
	new->previous = NULL;
	dll->current = dll->first = dll->last = new;
	result = SUCCESS;
    } else if (where == AFTER) {      
	new->next = dll->current->next;
	new->previous = dll->current;
	if (dll->current == dll->last)
	    dll->last = new;
	else
	    dll->current->next->previous = new;
	dll->current->next = new;    
	result = nextElement(dll);
    } else {
	new->next = dll->current;
	new->previous = dll->current->previous;
	if (dll->current == dll->first)
	    dll->first = new;
	else
	    dll->current->previous->next = new;
	dll->current->previous = new;
	result = prevElement(dll);
    }
    return result;
}

/**
   Deletes "current" element.
   @postcondition New "current" element is set the following element if it exists. If not, it is set to the previous.
   @return 0 on succes, -1 on failure
*/
int deleteElement(DLL *dll){
    int result = 0;
    if (isEmptyDLL(dll)) {
	result = FAILURE;
    } else if (nextElement(dll) == SUCCESS) {
	if (dll->current->previous == dll->first){ 
	    /*Deleting first element in list*/
	    dll->first = dll->current;
	    free(dll->current->previous);
	    dll->current->previous = NULL;
	} else {                                   
	    /*Deleting other element in list*/
	    dll->current->previous = dll->current->previous->previous;
	    free(dll->current->previous->next);
	    dll->current->previous->next = dll->current;
	}
    } else if (prevElement(dll) == SUCCESS){
	/*Deleting last element in list*/
	dll->last = dll->current;
	free(dll->current->next);
	dll->current->next = NULL;
    } else {                                       
	/*Only one element in list*/
	free(dll->current);
	dll->first = dll->last = dll->current = NULL;	
    }
    return result;
}

/**
   Applies the function "f" on all elements in the list.
   @postcondition "current" element is not changed.
   @return all return values from the function calls bitwise OR'ed (|) together.
*/
int mapDLL(DLL *dll, int (*f)(ELEMENT)){
    NODE *calc;
    int result = 0;
    calc = dll->first;
    do {
	result = result | f(calc->data);
	calc = calc->next;
    } while (calc != NULL);
    return result;
}

/**
   Searches the list and for ELEMENT using the equal function.
   The search terminates when the equal-function returns a non-zero value,
   indicating that the ELEMENT was found in the list.
   @postcondition "current" element is set to the found element. If not found then "current" element is not changed.
   @return 0 if an element was found, otherwise -1.
*/
int searchDLL(DLL *dll, ELEMENT element, int (*equal)(ELEMENT, ELEMENT)){
    NODE *search;
    NODE *found;
    int result = 0;
    search = dll->first;
    
    do {
	found = search;
	result = equal(element, search->data);
	search = search->next;
    } while ((result == 0) && (search != NULL));
    
    if (result != 0) {
	result = SUCCESS;
	dll->current = found;
    } else {
	result = FAILURE;
    }
    return result;
}

/**
   The list is sorted in descending order by comparing elements in the list with the cmp-function. The cmp-function must return < 0 if the first element is smaller than the second, == 0 if the elements are equal and > 0 otherwise.
   @postcondition "current" element is not changed.
   @return void.
*/
void sortDLL(DLL *dll, int (*cmp)(ELEMENT, ELEMENT)){

  if (dll->first != dll->last) {
    quicksort(dll, dll->first, dll->last, cmp);   
  }
}

/* The classical quicksort algorithm implemented. */
void quicksort(DLL *dll, NODE *p, NODE *r, int (*cmp)(ELEMENT, ELEMENT)){
  NODE *q;
  if ((p != r) && (p->previous != r)){
    q = partition(dll,p,r, cmp);
    quicksort(dll,p,q, cmp);
    quicksort(dll,q->next,r, cmp);
  }
}

NODE *partition(DLL *dll, NODE *p, NODE *r, int (*cmp)(ELEMENT, ELEMENT)){
  ELEMENT temp;
  NODE *i = (NODE *) malloc(sizeof(NODE));
  NODE *j = (NODE *) malloc(sizeof(NODE));
  int passed = FALSE;
  ELEMENT x = p->data;
  NODE *temp_i = i;
  NODE *temp_j = j;
  i->next = p;
  j->previous = r;

  while (!passed) {    
      /*Find element smaller than first element starting with last element */
    do {
      j = j->previous;
      if (j == i){
	passed = TRUE;
      }
    } while ((j != NULL) && (cmp(j->data, x) < 0));

    /*Find element larger than first element starting with first element */
    do {
      i = i->next;
      if (i == j){
	passed = TRUE;
      }
    } while (cmp(i->data, x) > 0);    
    
    if (!passed) {
        /*Switch element*/
      temp = i->data;
      i->data = j->data;
      j->data = temp;
      /*Restore current*/
      if (dll->current == i){
	dll->current = j;
      } else if (dll->current == j){
	dll->current = i;
      }
    }
  }
  free(temp_i);
  free(temp_j);
  return j;
}
